# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    imu.py                                             :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: taston <thomas.aston@ed.ac.uk>             +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2023/04/26 16:03:23 by taston            #+#    #+#              #
#    Updated: 2023/05/10 11:38:08 by taston           ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

from .sensordata import SensorData
import pandas as pd


class IMU(SensorData):
    """
    A class used to represent an IMU

    ...
    
    Attributes
    ----------
    filename : str
        file containing IMU sensor data
    id : int, float, str
        unique identifier given to IMU
    time_offset : float
        time offset applied to IMU data to sync with head pose data
    """
    counter = 0
    def __init__(self, filename=None, time_offset=0, id=counter):
        """
        Parameters
        ----------
        filename : str
            file containing IMU sensor data
        id : int, float, str
            unique identifier given to IMU
        time_offset : float
            time offset applied to IMU data to sync with head pose data
        """
        super().__init__()
        IMU.counter += 1
        self.filename = filename
        self.time_offset = time_offset
        if id:
            self.id = id
        else:
            self.id = IMU.counter


class Wax9(IMU):
    """
    A class representing a Wax9 IMU: https://axivity.com/downloads/wax9
    
    ...
    
    Attributes
    ----------
    columns : list
        list of sensor data column headings

    Methods
    -------
    extract_from_file()
        Extracts kinematic data from file provided
    """
    def __init__(self, filename, time_offset=0, id=False):
        """
        Parameters
        ----------
        filename : str
            file containing IMU sensor data
        id : int, float, str
            unique identifier given to IMU
        time_offset : float
            time offset applied to IMU data to sync with head pose data
        """
        super().__init__(filename, time_offset, id)
        self.columns = ['sensor', 
                        'received time','sample number','sample time',
                        'accelX','accelY','accelZ',
                        'gyroX','gyroY','gyroZ',
                        'magX','magY','magZ']
        self.extract_from_file()

    def extract_from_file(self):
        """
        Extracts kinematic data from file provided
        """
        data = pd.read_csv(self.filename)
        data.columns = self.columns
        data['adjusted time'] = data['sample time'] - data['sample time'][0] + self.time_offset

        self.velocity['time'] = data['adjusted time']
        self.velocity['yaw'] = data['gyroX']
        self.velocity['pitch'] = data['gyroY']
        self.velocity['roll'] = data['gyroZ']


# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    video.py                                           :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: taston <thomas.aston@ed.ac.uk>             +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2023/04/25 10:00:46 by taston            #+#    #+#              #
#    Updated: 2023/05/12 14:26:00 by taston           ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

import cv2

class Video:
    """
    A class representing a Video

    ...
    
    Attributes
    ----------
    cap : 

    filename : str
        path to video file
    fps : int
        framerate in frames per second
    height : int
        pixel height of video
    total_frames : int
        length of video in frames
    width : int
        pixel width of video
    writer : 


    Methods
    -------
    create_writer()
        creates video writer object
    get_dim()
        gets video dimensions
    get_fps()
        gets video fps
    get_length()
        gets length of video in frames
    """
    def __init__(self, filename=None):
        if filename:
            self.filename = filename
            self.cap = self._open_vid()
            self.width, self.height = self.get_dim()
            self.total_frames = self.get_length()
            self.fps = self.get_fps()

    def _open_vid(self):
        '''
        Open specified video file and create capture object
        '''
        cap = cv2.VideoCapture(self.filename)
        return cap
    
    def create_writer(self):
        '''
        Create writer object for recording videos based on chosen video.
        '''
        self.writer = cv2.VideoWriter('calibration.mp4',
                            cv2.VideoWriter_fourcc(*'mp4v'),
                            self.fps,
                            (self.width, self.height))

    def get_dim(self):
        '''
        Get video resolution
        '''
        width = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        return width, height

    def get_length(self):
        '''
        Get length of video in frames
        '''
        total_frames = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))
        return total_frames
    
    def get_fps(self):
        '''
        Get framerate of video in fps
        '''
        fps = round(self.cap.get(cv2.CAP_PROP_FPS))
        return fps
    
    def __str__(self):
        return ('-'*60 + '\n' + 
                'Video data:' + '\n' + 
                '-'*60 + '\n' + 
                f'Video resolution: {self.width} x {self.height} pixels' + '\n' +
                f'Length of video: {self.total_frames} frames' + '\n' + 
                f'Framerate: {self.fps} fps' + '\n' +
                '-'*60)
        # return 'Video'
    
    

    


from .camera import(
    Camera
)

from .video import(
    Video
)

from .calibration import (
    Calibrator
)

from .facedetector import(
    MediaPipe
)

from .imu import(
    Wax9
)

from .filter import(
    Filter
)

from .head import(
    Head
)

from .filter import(
    Filter
)

from .sensordata import(
    SensorData
)

from .plot import(
    Plot
)


from .resources import *

__all__ = [
    "calibrate",
    "log",
    "track",
    "filter",
    "process",
    "compare",
    "archive"
]


# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    camera.py                                          :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: taston <thomas.aston@ed.ac.uk>             +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2023/04/25 15:41:23 by taston            #+#    #+#              #
#    Updated: 2023/05/12 14:12:59 by taston           ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

import numpy as np
from .video import Video
from .calibration import Calibrator

class Camera:
    """
    A class used to represent a Camera.

    ...

    Attributes
    ----------
    focal_length : float
        float representing the focal length of the Camera in mm
    internal_matrix : ndarray
        array representing the Camera's intrinsic parameters
    distortion_matrix : ndarray
        array representing the Camera's lens distortion parameters
    calibrator : Calibrator
        Calibrator object used for camera calibration
    calibrated : bool
        bool for quick checking if camera has been calibrated
    video : Video
        Video object where the footage has been shot using this Camera
        
    Methods
    -------
    calibrate(checkerboard=(9,6), video=Video()):
        Performs calibration on the camera
    """
    def __init__(self):
        """
        Parameters
        ----------
        ...
        """

        width = 1920
        height = 1080
        self.focal_length = height * 1.28
        self.internal_matrix = np.array([[self.focal_length, 0, width/2],
                                          [0, self.focal_length, height/2],
                                          [0, 0, 1]])
        self.distortion_matrix = np.zeros((4, 1), dtype=np.float64)
        self.calibrated = False

    def calibrate(self, checkerboard=(9,6), video=Video()):
        """Creates a calibrator object and calibrates the Camera.

        If arguments checkerboard and video aren't passed in, the
        default checkerboard pattern and an empty video are used.

        Parameters
        ----------
        checkerboard : tuple, optional
            Checkerboard pattern used in camera calibration (default is 9x6)
        video : Video, optional
            Video used to calibrate camera
        """
        
        self.video = video
        self.calibrator = Calibrator(checkerboard, self.video)
        self.calibrated = True

        return self


# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    filter.py                                          :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: taston <thomas.aston@ed.ac.uk>             +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2023/04/26 15:59:05 by taston            #+#    #+#              #
#    Updated: 2023/05/10 11:29:07 by taston           ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

import scipy
import numpy as np


class Filter:
    """
    A class used to represent a Filter for data

    ...
    
    Attributes
    ----------
    a : ndarray
        denominator polynomial of the IIR filter
    b : ndarray
        numerator polynomial of the IIR filter

    More information on IIR filters here:
    https://en.wikipedia.org/wiki/Infinite_impulse_response

    Methods
    -------
    low_pass_butterworth(fs=4000, lowcut=160, order=4)
        creates a low pass butterworth filter
    apply(signal)
        applies filter to a given signal
    """
    def __init__(self):
        nyq = 2000
        low = 160 / nyq
        self.b, self.a = scipy.signal.butter(4, low, analog=False)
        
    def low_pass_butterworth(self, fs=4000, lowcut=160, order=4):
        """Creates a low pass butterworth filter

        Parameters
        ----------
        fs : float, optional
            sampling frequency (default 4000Hz)
        lowcut : float, optional
            lowcut frequency (default 160Hz)
        order : int, optional
            order of the filter (default 4)

        Returns
        -------
        self
        """
        nyq = 0.5 * fs
        low = lowcut / nyq
        
        # Create scipy Butterworth filter:
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.butter.html
        self.b, self.a = scipy.signal.butter(order, low, analog=False)
        
        return self
    
    def apply(self, signal):
        """Applies filter to a signal
        
        Parameters
        ----------
        signal : list
            signal for filter to be applied to
        
        Returns
        -------
        filtered_signal : list
            new filtered signal
        """
        filtered_signal = scipy.signal.filtfilt(self.b, self.a, signal[~np.isnan(signal)])
        return filtered_signal


import EdiHeadyTrack.logging
import matplotlib.pyplot as plt 


def test_log_show_false():
    assert EdiHeadyTrack.logging.log(show=False) == None

def test_log_show_true():
    assert EdiHeadyTrack.logging.log(show=True) == None


# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    plot.py                                            :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: taston <thomas.aston@ed.ac.uk>             +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2023/04/27 10:18:49 by taston            #+#    #+#              #
#    Updated: 2023/05/12 15:02:08 by taston           ###   ########.fr        #
#                                                                              #
# **************************************************************************** #


from .head import Head
from .imu import IMU
from .sensordata import SensorData
import matplotlib.pyplot as plt
from matplotlib.offsetbox import (TextArea, DrawingArea, OffsetImage,
                                  AnnotationBbox)

class Plot:
    """
    A class representing a Plot

    ...
    
    Attributes
    ----------
    colors : list
        list of colors used in plots

    Methods
    -------
    plot_comparison(property, xlim, ylim, key_times, *sensors)
        plots a time comparison between sensor data
    plot_head(property, *heads)
        plots head pose estimation data only
    plot_imu(property, time_range, *imus)
        plots IMU data only
    """
    def __init__(self):
        self.colors = []

    def plot_head(self, property, *heads):
        """Plots head pose estimation data only
        
        Parameters
        ----------
        property : str
            specifies which property to be plotted (from ('pose', 'velocity', 'acceleration')
        *heads : Head
            head objects to be plotted
        """
        plt.rcParams.update({'font.size': 14})
        fig = plt.figure('EdiHeadyTrack - Head Data Plotting') #,figsize=[7,7])    
        
        for head in heads:
            if not isinstance(head, Head):
                raise TypeError(f'Attempted to plot type {type(head)}. Only Head objects may be plotted here!')
            
            if hasattr(head, property):
                property_dict = getattr(head, property)
                for idx, key in enumerate(list(property_dict.keys())[1:]):
                    ax = plt.subplot(len(list(property_dict.keys()))-1, 1, idx+1)
                    ax.grid(color='0.95')
                    if property == 'pose':
                        ax.set_ylabel(r'$\theta$ (deg)')
                    elif property == 'velocity':
                        ax.set_ylabel(r'$\omega$ (deg/s)')
                    elif property =='acceleration':
                        ax.set_ylabel(r'$\alpha$ (deg/s$^2$)')

                    ax.plot(property_dict['time'], property_dict[key], label=f'{head.id} {key}')
                    ax.legend(loc='best')

        plt.xlabel(r'Time (s)')
        plt.tight_layout()

        plt.show()

    def plot_imu(self, property, time_range, *imus):
        """Plots head pose estimation data only
        
        Parameters
        ----------
        property : str
            specifies which property to be plotted (from ('pose', 'velocity', 'acceleration')
        time_range : tuple, float
            range of times to be plotted over (lower limit, upper limit)
        *imus : IMU
            IMU objects to be plotted
        """
        plt.rcParams.update({'font.size': 14})
        fig = plt.figure('EdiHeadyTrack - Head Data Plotting',figsize=[7,7])    

        for imu in imus:
            if not isinstance(imu, IMU):
                raise TypeError(f'Attempted to plot type {type(imu)}. Only IMU objects may be plotted here!')

            if hasattr(imu, property):
                property_dict = getattr(imu, property)
                for idx, key in enumerate(list(property_dict.keys())[1:]):
                    ax = plt.subplot(len(list(property_dict.keys()))-1, 1, idx+1)
                    ax.grid(color='0.95')
                    if property == 'pose':
                        ax.set_ylabel(r'$\theta$ (deg)')
                    elif property == 'velocity':
                        ax.set_ylabel(r'$\omega$ (deg/s)')
                    elif property =='acceleration':
                        ax.set_ylabel(r'$\alpha$ (deg/s$^2$)')

                    if time_range:
                        ax.set_xlim(time_range)

                    ax.plot(property_dict['time'], property_dict[key], label=f'{imu.id} {key}')
                    ax.legend(loc='best')

        plt.xlabel(r'Time (s)')
        plt.tight_layout()

        plt.show()


    def plot_comparison(self, property, xlim, ylim, key_times=[], *sensors):
        """Plots comparison between sensor data
        
        Parameters
        ----------
        property : str
            specifies which property to be plotted (from ('pose', 'velocity', 'acceleration')
        xlim : tuple, float
            range of times to be plotted over (lower limit, upper limit)
        ylim : tuple, float
            range of y values to be plotted over (lower limit, upper limit)
        key_times : list, float, optional
            key time points to display frame images for 
        *sensors : SensorData
            sensor data objects to be added to plot
        """
        plt.rcParams.update({'font.size': 14})
        fig = plt.figure('EdiHeadyTrack - Head Data Plotting',figsize=[6,8])    

        key_frames = [int(240*time) for time in key_times]

        for sensor in sensors:
            if not isinstance(sensor, SensorData):
                raise TypeError(f'Attempted to plot type {type(sensor)}. Only SensorData objects may be plotted here!')

            if hasattr(sensor, property):
                property_dict = getattr(sensor, property)
                for idx, key in enumerate(list(property_dict.keys())[1:]):
                    ax = plt.subplot(len(list(property_dict.keys())), 1, idx+2)
                    ax.grid(color='0.95')
                    if property == 'pose':
                        ax.set_ylabel(fr'$\theta_{{{key}}}$ (deg)')
                    elif property == 'velocity':
                        ax.set_ylabel(fr'$\omega_{{{key}}}$ (deg/s)')
                    elif property =='acceleration':
                        ax.set_ylabel(fr'$\alpha_{{{key}}}$ (deg/s$^2$)')

                    if xlim:
                        ax.set_xlim(xlim)
                    if ylim:
                        ax.set_ylim(ylim)

                    for time in key_times:
                        ax.axvline(x=time, color='green', ls=':')
                    
                    if isinstance(sensor, Head):
                        ax.plot(property_dict['time'], property_dict[key], label=f'{sensor.id}', color='c', ls='dashed')
                    elif isinstance(sensor, IMU):
                        ax.plot(property_dict['time'], property_dict[key], label=f'{sensor.id}', color='k')

                    # ax.set_ylim(min(property_dict[key] - abs(min(property_dict[key])/10)) , max(property_dict[key] + abs(max(property_dict[key])/10)))
                    ax.set_ylim(-220, 220)

                    if idx + 2 != len(list(property_dict.keys())):
                        plt.setp(ax.get_xticklabels(), visible=False)
                    else:
                        ax.set_xlabel(r'Time (s)')
                    
                    if idx + 2 == 3: 
                        handles, labels = ax.get_legend_handles_labels()
                        ax.legend(handles, labels, bbox_to_anchor=(1.0, 0.85), loc=2)

        ax1 = plt.subplot(411, sharex=ax)
        for time in key_times:
            ax1.axvline(x=time, color='green', ls=':')
        plt.setp(ax1.get_xticklabels(), visible=False)
        plt.setp(ax1.get_yticklabels(), visible=False)

        for sensor in sensors:
            if isinstance(sensor, Head):
                for idx, frame in enumerate(key_frames):
                    img = sensor.facedetector.tracking_frames[frame]

                    frame_index = sensor.facedetector.face2d['frame'].index(frame)
                    x_list = [pos[0] for pos in sensor.facedetector.face2d['all landmark positions'][frame_index]]
                    y_list = [pos[1] for pos in sensor.facedetector.face2d['all landmark positions'][frame_index]]
                    
                    top_bound = max(0, min(y_list[:]) - 200)
                    bottom_bound = min(sensor.facedetector.video.height, max(y_list[:]) + 200)
                    left_bound = max(0, min(x_list[:]) - 200)
                    right_bound = min(sensor.facedetector.video.width, max(x_list[:]) + 200)
                    
                    img = img[top_bound:bottom_bound, left_bound:right_bound, :]

                    imagebox = OffsetImage(img, zoom=0.08)
                    imagebox.image.axes = ax1

                    ab = AnnotationBbox(imagebox, [key_times[idx],0.5],
                                        xybox=(1, 1),
                                        xycoords='data',
                                        boxcoords="offset points",
                                        pad=0.01,
                                        bboxprops =dict(edgecolor='white')
                                        )

                    ax1.add_artist(ab)
        
        plt.savefig('EdiHeadyTrack/resources/comparison.png', bbox_inches='tight')
        plt.show()


# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    facedetector.py                                    :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: taston <thomas.aston@ed.ac.uk>             +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2023/02/10 16:43:13 by taston            #+#    #+#              #
#    Updated: 2023/05/10 11:29:21 by taston           ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

import cv2
import mediapipe as mp
import numpy as np
from numpy import genfromtxt
from tqdm import tqdm
from datetime import datetime
from .camera import Camera
from .video import Video
from .filter import Filter

class FaceDetector:
    """
    A class for representing a FaceDetector
    
    ...
    
    Attributes
    ----------
    camera : Camera, optional
        Camera object to be used with FaceDetector
    face2d : dict
        dict of detected face points in 2d
    face3d : list
        list of known 3d face points (from mesh model)
    """
    def __init__(self, video=Video(), camera=Camera()):
        self.camera = camera
        self.video = video
        self.face2d = {'time': [],
                       'frame': [],
                       'key landmark positions':    [],
                       'all landmark positions':    []}
        self.face3d = []

class MediaPipe(FaceDetector):
    """
    A class for representing a MediaPipe FaceDetector

    ...
    
    Attributes
    ----------
    drawSpec : DrawingSpec
        Drawing specifications for face mesh
    face3d : list
        list of known 3d face points (from mesh model)
    faceMesh : FaceMesh
        MediaPipe FaceMesh object 
    key_landmarks : list
        list of key landmark positions used for pose estimation
    maxFaces : int
        int number of maximum faces to be detected in video
    minDetectionCon : float
        float representing minimum detection confidence
    minTrackCon : float
        float representing minimum tracking confidence
    mpDraw : module
        MediaPipe drawing utilities
    mpFaceMesh : module
        MediaPipe face mesh utilities
    refineLandmarks : bool
        bool for choosing if landmarks will be refined
    staticMode : bool
        bool representing if searching for landmarks on static frame
        or non static video file
    tracking_frames : list
        list of frame numbers which have successfully been tracked
        
    Methods
    -------
    find_faces(img)
        search for faces in a given frame
    run()
        run through the tracking procedure using MediaPipe face mesh
    """
    def __init__(self, video=Video(), camera=Camera(),
                 staticMode=False, maxFaces=1, refineLandmarks=False, minDetectionCon=0.5, minTrackCon=0.5):
        """
        Parameters
        ----------
        video : Video, optional
            video to perform head pose estimation on (default is empty Video)
        camera : Camera, optional
            camera used to capture video (default is uncalibrated Camera)
        staticMode : bool, optional
            flag specifying if tracking is static or not (default False)
        maxFaces : int, optional
            maximum number of faces allowed to be detected in video (default 1)
        refineLandmarks : bool, optional
            flag setting if landmarks should be refined or not (default False)
        minDetectionCon : float, optional
            minimum detection confidence (default 0.5)
        minTrackCon : float, optional
            minimum tracking confidence (default 0.5)
        """
        super().__init__(video, camera)
        timestamp = datetime.now().strftime("%H:%M:%S")
        print('-'*120)
        print('{:<100} {:>19}'.format(f'Creating MediaPipe object for video {self.video.filename}:', timestamp))
        print('-'*120)
        print(self.video)
        self.staticMode = staticMode
        self.refineLandmarks = refineLandmarks
        self.maxFaces = maxFaces
        self.minDetectionCon = minDetectionCon
        self.minTrackCon = minTrackCon
        self.mpDraw = mp.solutions.drawing_utils
        self.mpFaceMesh = mp.solutions.face_mesh
        print(type(self.mpDraw), type(self.mpFaceMesh))
        self.faceMesh = self.mpFaceMesh.FaceMesh(self.staticMode,
                                                 self.maxFaces,
                                                 self.refineLandmarks,
                                                 self.minDetectionCon,
                                                 self.minTrackCon)
        self.drawSpec = self.mpDraw.DrawingSpec(thickness=1, circle_radius=2)
        self.key_landmarks = [33, 263, 1, 61, 291, 199]
        self.run()
        timestamp = datetime.now().strftime("%H:%M:%S")
        print('-'*120)
        print('{:<100} {:>19}'.format(f'MediaPipe object complete!', timestamp))
        print('-'*120)
    
    def run(self):
        """Runs through MediaPipe tracking process, calling relevant
        functions
        
        More information found here:

        https://github.com/google/mediapipe/blob/master/docs/solutions/face_mesh.md
        """
        print('Running MediaPipe Face Mesh on selected video...')
        self.face3d = [[0, -1.126865, 7.475604], # 1
                       [-4.445859, 2.663991, 3.173422], # 33
                       [-2.456206,	-4.342621, 4.283884], # 61
                       [0, -9.403378, 4.264492], # 199
                       [4.445859, 2.663991, 3.173422], # 263
                       [2.456206, -4.342621, 4.283884]] # 291
        
        progress_bar = tqdm(range(self.video.total_frames))
        self.tracking_frames = []
        while True:
            success, img = self.video.cap.read()
            if success:
                progress_bar.update(1)
                # current_frame = int(self.video.cap.get(1))
                self.find_faces(img)
                cv2.namedWindow("EdiHeadyTrack", cv2.WINDOW_NORMAL)
                cv2.resizeWindow("EdiHeadyTrack", 640, 360)
                cv2.imshow("EdiHeadyTrack", img)
                # cv2.imwrite(f'tracking frames/{current_frame}.png', img)
                
                if cv2.waitKey(5) & 0xFF == ord('q'):
                    progress_bar.close()
                    print('Face tracking interuppted...')
                    break     
            else:
                progress_bar.close()
                print('Face tracking complete...')
                break

    def find_faces(self, img):
        """Finds faces in a supplied image

        Parameters
        ----------
        img : ndarray
            ndarray representing the image in which faces should be found
        """
        imgRGB = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        results = self.faceMesh.process(imgRGB)
        frame_number = int(self.video.cap.get(cv2.CAP_PROP_POS_FRAMES))
        time = frame_number / self.video.fps
        
        if results.multi_face_landmarks:
            self.face2d['time'].append(time)
            self.face2d['frame'].append(frame_number)
            landmark_positions=[]
            key_landmark_positions=[]
            for faceLandmarks in results.multi_face_landmarks:
                self.mpDraw.draw_landmarks(img,
                                           faceLandmarks,
                                           self.mpFaceMesh.FACEMESH_TESSELATION,
                                           None,
                                           mp.solutions.drawing_styles
                                           .get_default_face_mesh_tesselation_style())
                
                for idx, lm in enumerate(faceLandmarks.landmark):
                    x = int(lm.x * self.video.width)
                    y = int(lm.y * self.video.height)

                    landmark_position = [x,y]
                    landmark_positions.append(landmark_position)
                    
                    if idx in self.key_landmarks:
                        if idx == 1:
                            nose2d = (lm.x * self.video.width, lm.y * self.video.height)
                            nose3d = (lm.x * self.video.width, lm.y * self.video.height, lm.z * 3000)

                        key_landmark_positions.append(landmark_position)
                
                self.face2d['all landmark positions'].append(landmark_positions)
                self.face2d['key landmark positions'].append(key_landmark_positions)

        self.tracking_frames.append(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
        return
    
    def __str__(self):
        return f'MediaPipe Face Detector with video {self.video.filename}'


# class OpenPose(FaceDetector):
#     def __init__(self):
#         ...

# class FaceMeshDetector():
#     def __init__(self, staticMode=False, maxFaces=2, refineLandmarks=False, minDetectionCon=0.5, minTrackCon=0.5):
#         '''
#         Initialise object
#         '''
#         self.staticMode = staticMode
#         self.maxFaces = maxFaces
#         self.refineLandmarks = refineLandmarks
#         self.minDetectionCon = minDetectionCon
#         self.minTrackCon = minTrackCon

#         self.mpDraw = mp.solutions.drawing_utils
#         self.mpFaceMesh = mp.solutions.face_mesh
#         self.faceMesh = self.mpFaceMesh.FaceMesh(self.staticMode,
#                                                  self.maxFaces,
#                                                  self.refineLandmarks,
#                                                  self.minDetectionCon,
#                                                  self.minTrackCon)
#         self.drawSpec = self.mpDraw.DrawingSpec(thickness=1, circle_radius=2)

#     def findFaceMesh(self, img, draw=True):
#         '''
#         Finds face mesh and draws it
#         '''
#         img.flags.writeable = False
#         self.imgRGB = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
#         ih, iw, ic = img.shape
#         self.results = self.faceMesh.process(self.imgRGB)
#         face = {
#                     "x": [],
#                     "y": [],
#                     "z": [],
#                     "x_av": None,
#                     "y_av": None,
#                     "yaw": None,
#                     "pitch": None,
#                     "roll": None
#                 }
#         p1 = (0, 0)
#         p2 = (0, 0)
#         face_2d = []
#         face_3d = [[0, -1.126865, 7.475604], # 1
#                    [-4.445859, 2.663991, 3.173422], # 33
#                    [-2.456206,	-4.342621, 4.283884], # 61
#                    [0, -9.403378, 4.264492], # 199
#                    [4.445859, 2.663991, 3.173422], # 263
#                    [2.456206, -4.342621, 4.283884]] # 291

#         key_landmarks = [33, 263, 1, 61, 291, 199]
#         # If something detected
#         if self.results.multi_face_landmarks:
#             faces = []
#             for faceLandmarks in self.results.multi_face_landmarks:
#                 # self.mpDraw.draw_landmarks(img, faceLandmarks, self.mpFaceMesh.FACEMESH_TESSELATION,
#                 #                             self.drawSpec,self.drawSpec)
#                 self.mpDraw.draw_landmarks(img, 
#                                            faceLandmarks, 
#                                            self.mpFaceMesh.FACEMESH_TESSELATION,
#                                            None,
#                                            mp.solutions.drawing_styles
#                                            .get_default_face_mesh_tesselation_style())

                
#                 # Get landmarks
#                 for idx, lm in enumerate(faceLandmarks.landmark):
                    
#                     face["x"].append(int(lm.x*iw))
#                     face["y"].append(int(lm.y*ih))  
#                     face["z"].append(lm.z)
#                     if idx in key_landmarks:
#                         if idx == 1:
#                             nose_2d = (lm.x * iw, lm.y * ih)
#                             nose_3d = (lm.x * iw, lm.y*ih, lm.z * 3000)

#                         x, y = int(lm.x * iw), int(lm.y * ih)
#                         face_2d.append([x, y])
#                         # face_3d.append([x, y, lm.z])

#                 faces.append(face)
#                 # face_2d = [face["x"], face["y"]]
#                 # face_3d = [face["x"], face["y"], face["z"]]
#                 face_2d = np.array(face_2d, dtype=np.float64)
#                 face_3d = np.array(face_3d, dtype=np.float64)
#                 # print(face_2d)
#                 # print(face_3d)
#                 # Camera properties
#                 # camera_matrix = genfromtxt('camera_matrix.csv', delimiter=',')
#                 # print(camera_matrix)
#                 # distortion_matrix = genfromtxt('camera_distortion.csv', delimiter=',')
                
#                 focal_length = ih * 1.28
#                 camera_matrix = np.array([[focal_length, 0, iw/2],
#                                           [0, focal_length, ih/2],
#                                           [0, 0, 1]])
#                 distortion_matrix = np.zeros((4, 1), dtype=np.float64)
    
#                 # Solve PnP
#                 success, rot_vec, trans_vec = cv2.solvePnP(face_3d, face_2d, camera_matrix, distortion_matrix, flags=cv2.SOLVEPNP_ITERATIVE)
                    
#                 # Rotational matrix
#                 rmat = cv2.Rodrigues(rot_vec)[0]

#                 # Get angles
#                 P = np.hstack((rmat, np.zeros((3, 1), dtype=np.float64)))
#                 eulerAngles =  cv2.decomposeProjectionMatrix(P)[6]
#                 yaw   = eulerAngles[1, 0]
#                 pitch = eulerAngles[0, 0]
#                 roll  = eulerAngles[2, 0] 
                
#                 # Angle correction performed to centre around zero
#                 if pitch < 0:
#                     pitch = - 180 - pitch
#                 elif pitch >= 0: 
#                     pitch = 180 - pitch

#                 # Define points for direction line drawn on face
#                 if nose_2d:
#                     p1 = (int(nose_2d[0]), int(nose_2d[1]))
#                     p2 = (int(nose_2d[0] - yaw * 2), int(nose_2d[1] - pitch * 2))

        
#         else:
#             faces = None 
            
#         if faces:
#             for face in faces:
#                 if face["x"]:
#                     face["x_av"] = np.average(face["x"])
#                     face["y_av"] = np.average(face["y"])
#                     face["yaw"] = yaw
#                     face["pitch"] = pitch
#                     face["roll"] = roll
        
        
#         return img, face, faces, p1, p2
    



import EdiHeadyTrack.calibrating    # for calibrating camera, extracting intrinsic parameters

TEST_FILE = 'videos/calibration_xiaomi.mp4'

def test_initialise():
    '''
    Assert correct values are returned for dummy checkerboard dimensions
    '''
    checkerboard = (2,2)
    criteria, threedpoints, twodpoints, objectp3d = EdiHeadyTrack.calibrating.initialise(checkerboard)
    assert criteria == (3, 30, 0.001)
    assert threedpoints == []
    assert twodpoints == []
    import numpy as np
    assert np.array_equal(objectp3d, [[[0., 0., 0.],[1., 0., 0.],[0., 1., 0.],[1., 1., 0.]]])


def test_vid():
    '''
    Assert frames of video file can be read, and that video fps is greater than 0
    '''
    assert EdiHeadyTrack.calibrating.open_vid(TEST_FILE)[0].read()[0] == True
    assert EdiHeadyTrack.calibrating.open_vid(TEST_FILE)[1] > 0


# def test_calibrate():
#     '''
#     Assert that calibrate function runs without error
#     '''
#     assert EdiHeadyTrack.calibrating.calibrate(TEST_FILE) == None

# def test_webcam():
#     '''
#     Assert frames from webcam can be read, and that video fps is greater than 0
#     '''
#     assert EdiHeadyTrack.calibrating.open_vid(0)[0].read()[0] == True
#     assert EdiHeadyTrack.calibrating.open_vid(0)[1] > 0


# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    calibration.py                                     :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: taston <thomas.aston@ed.ac.uk>             +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2023/04/25 10:20:08 by taston            #+#    #+#              #
#    Updated: 2023/05/12 14:09:34 by taston           ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

import cv2 
import numpy as np
from datetime import datetime
from .video import Video

class Checkerboard:
    """
    A class used to represent a calibration Checkerboard

    ...
    
    Attributes
    ----------
    dimensions : tuple
        tuple of checkerboard pattern dimensions
    min_points : int
        integer threshold of minimum detected points for 
        checkerboard to be considered found
    objectp3d : ndarray
        array of checkerboard points in three dimensions
    threedpoints : list
        list of checkerboard points in three dimensions
        for each frame where a checkerboard is found
    twodpoints : list 
        list of detected checkerboard points in two
        dimensions for each frame
        
    Methods
    -------
    get_corners(gray_frame)
        Finds checkerboard corners in a given grayscale frame
    """
    def __init__(self, dimensions = (9,6)):
        """
        Parameters
        ----------
        dimensions : tuple, optional
            Checkerboard pattern used in camera calibration (default is 9x6)
        """
        
        print('Checkerboard created')
        self.dimensions = dimensions
        self.min_points = 50
        self.twodpoints = [] 
        self.threedpoints = [] 
        self.objectp3d = np.zeros((1, self.dimensions[0]
                            * self.dimensions[1],
                            3), np.float32)
        self.objectp3d[0, :, :2] = np.mgrid[0:self.dimensions[0],
                                    0:self.dimensions[1]].T.reshape(-1, 2)
        
    def get_corners(self, gray_frame):
        """
        Looks for checkerboard corners in a given grayscale video
        frame.

        Parameters
        ----------
        gray_frame : ndarray
            ndarray representing grayscale frame from video

        Returns
        -------
        ret : bool
            bool representing if corner search was successful
        corners : ndarray
            ndarray containing coordinates of corners
        """

        ret, corners = cv2.findChessboardCorners(
                            gray_frame, self.dimensions,
                            cv2.CALIB_CB_ADAPTIVE_THRESH
                            + cv2.CALIB_CB_FAST_CHECK +
                            cv2.CALIB_CB_NORMALIZE_IMAGE)

        return ret, corners

class Calibrator:
    """
    A class used to represent a camera Calibrator

    ...
    
    Attributes
    ----------
    checkerboard : Checkerboard

    criteria : tuple
        tuple of criteria for successful camera calibration
    distortion : ndarray
        ndarray of distortion parameters
    frame : ndarray
        ndarray representing video frame
    gray_frame : ndarray
        ndarray representing grayscale video frame
    matrix : ndarray
        ndarray representing camera intrinsic matrix
    r_vecs : ndarray
        ndarray of rotational vectors
    t_vecs : ndarray
        ndarray of translation vectors
        
    Methods
    -------
    calibrate()
        Perform camera calibration process
    draw_corners(corners)
        Draw checkerboard corners on video frame
    save_outputs()
        Save camera parameters to csv files
    """

    def __init__(self, checkerboard, video=Video()):
        """
        Parameters
        ----------
        checkerboard : tuple
            tuple representing Checkerboard pattern
        video : Video, optional
            Video used for camera calibration. If no video specified
            an empty video will be attempted.
        """
        timestamp = datetime.now().strftime("%H:%M:%S")
        self.video = video
        print('-'*120)
        print('{:<100} {:>19}'.format(f'Creating Calibrator object for video {self.video.filename}:', timestamp))
        print('-'*120)
        print(self.video)
        self.checkerboard = Checkerboard(checkerboard)
        print(f'Checkerboard dimensions: {self.checkerboard.dimensions[0]} x {self.checkerboard.dimensions[1]}')
        self.criteria = (cv2.TERM_CRITERIA_EPS + 
                         cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
        self.calibrate()
        # self.save_outputs()
        timestamp = datetime.now().strftime("%H:%M:%S")
        print('-'*120)
        print('{:<100} {:>19}'.format(f'Calibrator object complete!', timestamp))
        print('-'*120)

    def calibrate(self):
        """
        Performs the camera calibration procedure outlined here:

        https://docs.opencv.org/4.x/dc/dbb/tutorial_py_calibration.html
        """
        timestamp = datetime.now().strftime("%H:%M:%S")
        self.video.create_writer()
        print('Displaying video...')
        while True:
            ret, self.frame = self.video.cap.read()
            frame_number = int(self.video.cap.get(cv2.CAP_PROP_POS_FRAMES))
            self.gray_frame = cv2.cvtColor(self.frame, cv2.COLOR_BGR2GRAY)
            ret, corners = self.checkerboard.get_corners(self.gray_frame)
            if ret:
                complete, image = self.draw_corners(corners)
            if complete: break
            cv2.imshow('Calibrating...', self.frame)
            self.video.writer.write(self.frame)
            k = cv2.waitKey(1)
            if k == 27:
                self.video.cap.release()
                self.video.writer.release()
                cv2.destroyAllWindows()
                break
        h, w = image.shape[:2] 
    
        # Perform camera calibration by given threedpoints and twodpoints
        ret, self.matrix, self.distortion, self.r_vecs, self.t_vecs = cv2.calibrateCamera(self.checkerboard.threedpoints, 
                                                                      self.checkerboard.twodpoints,
                                                                      self.gray_frame.shape[::-1], None, None)
        print(f'Number of frames used for calibration: {frame_number}')

        return self
    
    def draw_corners(self, corners):
        '''
        Draws corners of checkerboard onto frame to verify calibration is working

        Parameters
        ----------
        corners : ndarray
            ndarray of the corners found for a given frame

        Returns
        -------
        complete : bool
            bool representing whether search for corners is complete
        frame : ndarray
            new video frame with corners drawn
        '''
        complete = False
        
        self.checkerboard.threedpoints.append(self.checkerboard.objectp3d)
        # Refining pixel coordinates for given 2d points.
        corners2 = cv2.cornerSubPix(
            self.gray_frame, corners, self.checkerboard.dimensions, (-1, -1), self.criteria)
        self.checkerboard.twodpoints.append(corners2)
        # When we have minimum number of data points, stop:
        if len(self.checkerboard.twodpoints) > self.checkerboard.min_points:
            self.video.cap.release()
            self.video.writer.release()
            cv2.destroyAllWindows()
            complete=True

        # Draw and display the corners:
        frame = cv2.drawChessboardCorners(self.frame,
                                        self.checkerboard.dimensions,
                                        corners2, True)
            
        return complete, frame
    
    def save_outputs(self):
        """
        Saves matrices to csv
        """
        timestamp = datetime.now().strftime("%H:%M:%S")
        print('Saving outputs...')
        from numpy import savetxt
        savetxt('EdiHeadyTrack/resources/camera_matrix.csv', self.matrix, delimiter=',')
        savetxt('EdiHeadyTrack/resources/camera_distortion.csv', self.distortion, delimiter=',')


# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    head.py                                            :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: taston <thomas.aston@ed.ac.uk>             +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2023/04/26 11:14:49 by taston            #+#    #+#              #
#    Updated: 2023/05/12 14:40:15 by taston           ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

import cv2
import numpy as np
from datetime import datetime
import pandas as pd
from .facedetector import FaceDetector
from .filter import Filter
from .sensordata import SensorData

class Head(SensorData):
    """
    A class used to represent a Head obtained using head pose estimation
    techniques

    ...
    
    Attributes
    ----------
    facedetector : FaceDetector
        the FaceDetector object used to find the Head
    filter : Filter
        Filter object used for data filtering
    id : str, int, float
        unique identifier given to Head
    pose : dict
        dict containing Head pose time history

    Methods
    -------
    apply_filter(filter)
        applies filter to head pose data
    calculate_kinematics()
        calculates kinematic data from pose time history
    calculate_pose()
        computes HeadPose from detected facial landmarks
    """
    counter = 0
    def __init__(self, facedetector=FaceDetector(), id=counter):
        """
        Parameters
        ----------
        facedetector : FaceDetector, optional
            the FaceDetector object used to find the Head (default FaceDetector())
        id : str, int, float, optional
            unique identifier for each Head object (defaults to count of existing heads)
        """
        timestamp = datetime.now().strftime("%H:%M:%S")
        print('-'*120)
        print('{:<100} {:>19}'.format(f'Creating Head object for {facedetector}:', timestamp))
        print('-'*120)
        super().__init__()
        Head.counter += 1
        self.facedetector = facedetector
        # if id:
        self.id = id
        # # else:
        #     self.id = id
        self.pose = {'time':    [],
                     'yaw':     [],
                     'pitch':   [],
                     'roll':    []}
        
        self.calculate_pose()
        self.calculate_kinematics()
        timestamp = datetime.now().strftime("%H:%M:%S")
        print('-'*120)
        print('{:<100} {:>19}'.format(f'Head object complete!', timestamp))
        print('-'*120)

    def apply_filter(self, filter=Filter()):
        """Applies filter to head pose data and updates pose 
        
        Parameters
        ----------
        filter : Filter, optional
            Filter object used to filter data (default Filter())

        Returns
        -------
        self
        """
        # print('Filtering data...')
        self.filter = filter
        data = pd.DataFrame.from_dict(self.pose)
        properties = ['yaw', 'pitch', 'roll']
        for property in properties:
            signal = np.array(data[property])
            filtered_signal = self.filter.apply(signal)
            self.pose[property] = list(filtered_signal)
        
        self.calculate_kinematics()
        
        return self
    
    def calculate_pose(self):
        """Calculates head pose from detected facial landmarks using 
        Perspective-n-Point (PnP) pose computation:
        
        https://docs.opencv.org/4.6.0/d5/d1f/calib3d_solvePnP.html
        """
        print('Computing head pose from tracking data...')
        for idx, time in enumerate(self.facedetector.face2d['time']):
            self.pose['time'].append(time)
            face2d = self.facedetector.face2d['key landmark positions'][idx]
            face2d = np.array(face2d, dtype=np.float64)
            face3d = np.array(self.facedetector.face3d, dtype=np.float64)

            success, rot_vec, trans_vec = cv2.solvePnP(face3d,
                                                       face2d,
                                                       self.facedetector.camera.internal_matrix,
                                                       self.facedetector.camera.distortion_matrix,
                                                       flags=cv2.SOLVEPNP_ITERATIVE)
            
            rmat = cv2.Rodrigues(rot_vec)[0]

            P = np.hstack((rmat, np.zeros((3, 1), dtype=np.float64)))
            eulerAngles =  cv2.decomposeProjectionMatrix(P)[6]
            yaw = eulerAngles[1, 0]
            pitch = eulerAngles[0, 0]
            roll = eulerAngles[2,0]
            
            if pitch < 0:
                pitch = - 180 - pitch
            elif pitch >= 0: 
                pitch = 180 - pitch
            
            self.pose['yaw'].append(yaw)
            self.pose['pitch'].append(pitch)
            self.pose['roll'].append(roll)

    def calculate_kinematics(self):
        """
        Calculates kinematic data from pose time history
        """
        # print('Calculating kinematic data from head pose...')

        self.velocity['time'] = self.pose['time'][1:]
        self.acceleration['time'] = self.pose['time'][2:]

        for key in list(self.pose.keys())[1:]:
            self.velocity[key] = np.diff(np.array(self.pose[key])) / np.diff(np.array(self.pose['time']))
            self.acceleration[key] = np.diff(np.array(self.velocity[key])) / np.diff(np.array(self.velocity['time']))





    
    